/* ==========================================================================
 * libevnet/src/nio.c - Network server library for libevent.
 * --------------------------------------------------------------------------
 * Copyright (c) 2006  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */
#include <stdio.h>
#include <stdlib.h>	/* NULL */

#include <string.h>	/* strerror(3) memchr(3) memmove(3) */

#include <errno.h>	/* EAGAIN */

#include <assert.h>	/* assert(3) */

#include <unistd.h>	/* read(2) write(3) */

#include <sys/types.h>
#include <sys/param.h>	/* MAX */
#include <sys/queue.h>	/* SLIST_ENTRY SLIST_HEAD SLIST_INIT SLIST_FIRST
			 * SLIST_NEXT */
#include <sys/time.h>	/* struct timeval */

#include <err.h>

#include <event.h>

#include <arena/proto.h>

#include "nio.h"


#define MARK	fprintf(stderr,"@line %d\n",__LINE__);


#ifndef NIO_RECURSE_MAX
#define NIO_RECURSE_MAX	(3)
#endif

unsigned int nio_recurse_max	= NIO_RECURSE_MAX;


const char *nio_errlist[] = {
	[NIO_ESUCCESS]		= "Success",
	[NIO_ESYSTEM]		= "System error",
	[NIO_ETIMEDOUT]		= "Operation timed out",
	[NIO_EPENDING]		= "Operation already pending",
	[NIO_ECANCELLED]	= "Operation cancelled",
	[NIO_ENOMORE]		= "End of IO stream",
	[NIO_ENOSPACE]		= "No buffer space available",
	[NIO_ENODATA]		= "No more data is available on the write queue",
}; /* nio_errlist[] */

const int nio_nerr	= sizeof nio_errlist / sizeof nio_errlist[0];



/*
 * During possibly recursive calls keep stack state (or automatic storage)
 * within this structure, which should be pushed onto the stack data queue
 * of the local nio object. If the nio object is closed, we can signal this
 * to all the functions that are in scope.
 *
 * This allows information to be relayed safely across the stack and so
 * remove any dependency on the nio object existing for the entire
 * duration of any running function call (thus allowing them to do useful
 * stuff, like call optimizations, after calling the local callback).
 *
 * Without this we could not implement tail recursion. Note that a
 * longjmp(3) would really throw a monkey wrench into things.
 */
struct nio_stack_data {
	struct nio *nio;

	void *fn_name;

	SLIST_ENTRY(nio_stack_data) sle;
};


struct nio_call {
	void *fn_name;

	union {
		struct {
			void *cb_arg;

			struct timeval expire;
		} rdwr;

		struct {
			void *cb_arg;

			struct timeval expire;

			unsigned char *dst;

			size_t dstlen;

			unsigned char *dstpos;

			nio_read_cb cb;
		} read;

		struct {
			void *cb_arg;

			struct timeval expire;

			const unsigned char *src;

			size_t srclen;

			const unsigned char *srcpos;

			nio_write_cb cb;
		} write;

		struct {
			void *cb_arg;

			struct timeval expire;
			struct timeval timeout;

			nio_getln_cb cb;

			struct {
				char *top;
				char *pos;
				char *end;
			} buf;
		} getln;
	} fn_args;
};


static const struct nio {
	const struct arena_prototype *ap;

	struct event_base *ev_base;

	struct nio_options opts;

	int fd;

	enum nio_errno our_errno;
	int sys_errno;

	struct nio_type {
		struct nio_call call;
		struct nio_call *next; /* Pointer to .call. */

		SLIST_HEAD(, nio_stack_data) call_stack;
		SLIST_HEAD(, nio_stack_data) note_stack;

		unsigned int num_calls;

		struct event ev;

		int ev_pending;
		int ev_mode;
	} rd, wr, gl;
} nio_initializer;


static void nio_rdwr_try(struct nio *, struct nio_type *, int);

static void nio_read_catch(int fd, short events, void *nio) {
	if (events & EV_TIMEOUT)
		((struct nio *)nio)->rd.ev_pending	= 0;
	/* else check if we set EV_PERSIST */

	nio_rdwr_try(nio, &((struct nio *)nio)->rd, 0);

	return /* void */;
} /* nio_read_catch() */


static void nio_write_catch(int fd, short events, void *nio) {
	if (events & EV_TIMEOUT)
		((struct nio *)nio)->wr.ev_pending	= 0;
	/* else check if we set EV_PERSIST */

	nio_rdwr_try(nio, &((struct nio *)nio)->wr, 0);

	return /* void */;
} /* nio_write_catch() */


static void nio_rdwr_try(struct nio *nio, struct nio_type *t, int queue_only) {
	struct nio_stack_data sd	= { .nio = nio, .fn_name = &nio_rdwr_try };
	struct nio_call *cp	= &t->call;
	struct timeval now		= { 0 };
	struct nio_stack_data *sdp;
	struct timeval timeout;
	enum nio_errno err;
	enum nio_mode m;
	ssize_t n;

	/*
	 * Cancel out other calls on this stack.
	 */
	while ((sdp = SLIST_FIRST(&t->note_stack))) {
		SLIST_REMOVE_HEAD(&t->note_stack, sle);

		sdp->nio	= 0;
	}

	SLIST_INSERT_HEAD(&t->note_stack, &sd, sle);

	if (queue_only) {	/* Fake a blocking call. */
		n	= -1;
		errno	= EAGAIN;
	} else if (t == &nio->rd) {
		n	= nio->opts.sys_read(nio, nio->fd, cp->fn_args.read.dstpos, &cp->fn_args.read.dst[cp->fn_args.read.dstlen] - cp->fn_args.read.dstpos, cp->fn_args.read.cb_arg);
	} else {
		n	= nio->opts.sys_write(nio, nio->fd, cp->fn_args.write.srcpos, &cp->fn_args.write.src[cp->fn_args.write.srclen] - cp->fn_args.write.srcpos, cp->fn_args.write.cb_arg);
	}

	if (n == -1 && errno == EAGAIN) {
		if (0 != gettimeofday(&now,0)) {
			nio->sys_errno	= errno;
			nio->our_errno	= err
					= NIO_ESYSTEM;

			m	= NIO_ABORT;
		} else if (timercmp(&now, &cp->fn_args.rdwr.expire, >=)) {
			nio->our_errno	= err
					= NIO_ETIMEDOUT;

			m	= NIO_ABORT;
		} else {
			/*
			 * Query the IO mode/direction that we need to poll
			 * on before attempting our next IO operation. This
			 * is for protocols like SSL where, in order to
			 * fulfill a logical read, it might need to complete
			 * a low-level write, or vice versa.
			 */
			if (NIO_ABORT == (m = nio->opts.sys_mode(nio, nio->fd, ((t == &nio->rd)? NIO_READ : NIO_WRITE), t->call.fn_args.rdwr.cb_arg))) {
				nio->our_errno	= err
						= NIO_ECANCELLED;
			}
		}

		n	= 0;
	} else if (n < 0) {
		nio->sys_errno	= errno;
		nio->our_errno	= err
				= NIO_ESYSTEM;

		n	= 0;
		m	= NIO_ABORT;
	} else if (n == 0) {
		nio->our_errno	= err
				= NIO_ENOMORE;

		m	= NIO_ABORT;
	} else if (t == &nio->rd) {
		void *fn_name	= cp->fn_name;
		void *pos	= cp->fn_args.read.dstpos;
		int again;

		cp->fn_name		= 0;
		cp->fn_args.read.dstpos	+= n;

		if ((again = cp->fn_args.read.cb(sd.nio, pos, n, NIO_ESUCCESS, cp->fn_args.read.cb_arg))) {
			assert(sd.nio && !t->call.fn_name);

			/* Restart IO. */
			cp->fn_name	= fn_name;
			n		= 0;

			if (cp->fn_args.read.dstpos >= &cp->fn_args.read.dst[cp->fn_args.read.dstlen]) {
				m	= NIO_ABORT;

				nio->our_errno	= err
						= NIO_ENOSPACE;
			} else if (NIO_ABORT == (m = nio->opts.sys_mode(nio, nio->fd, NIO_READ, t->call.fn_args.read.cb_arg))) {
				nio->our_errno	= err
						= NIO_ECANCELLED;
			}
		} else if (sd.nio) {
			/* Do not allow another callback. */
			cp	= 0;

			/* Remove the event. */
			m	= NIO_ABORT;
		} else
			goto done;
	} else if (t == &nio->wr) {
		void *fn_name	= cp->fn_name;
		const void *pos	= cp->fn_args.write.srcpos;
		int again;

		cp->fn_name			= 0;
		cp->fn_args.write.srcpos	+= n;

		if ((again = cp->fn_args.write.cb(sd.nio, pos, n, NIO_ESUCCESS, cp->fn_args.write.cb_arg))) {
			assert(sd.nio && !sd.nio->wr.call.fn_name);

			/* Restart IO. */
			cp->fn_name	= fn_name;
			n		= 0;

			if (cp->fn_args.write.srcpos >= &cp->fn_args.write.src[cp->fn_args.write.srclen]) {
				m	= NIO_ABORT;

				nio->our_errno	= err
						= NIO_ENODATA;
			} else if (NIO_ABORT == (m = nio->opts.sys_mode(nio, nio->fd, NIO_WRITE, t->call.fn_args.rdwr.cb_arg))) {
				nio->our_errno	= err
						= NIO_ECANCELLED;
			}
		} else if (sd.nio) {
			/* Do not allow another callback. */
			cp	= 0;

			/* Remove the event. */
			m	= NIO_ABORT;
		} else
			goto done;
	} else {
		assert(0 && (t == &nio->rd || t == &nio->wr));
	}

	switch (m) {
	case NIO_READ:
		/* FALL THROUGH */
	case NIO_WRITE:
		if (t->ev_pending) {
			if (t->ev_mode == m)
				break;
 
			(void)event_del(&t->ev);

			t->ev_pending	= 0;
		}

		event_set(&t->ev, sd.nio->fd, ((m == NIO_READ)? EV_READ : EV_WRITE) | EV_PERSIST, (t == &sd.nio->rd)? &nio_read_catch : &nio_write_catch, nio);

		if (sd.nio->ev_base)
			event_base_set(sd.nio->ev_base,&t->ev);

		if (0 != gettimeofday(&now,0)) {
			nio->sys_errno	= errno;
			nio->our_errno	= err
					= NIO_ESYSTEM;

			goto abort;
		}

		if (timercmp(&now, &cp->fn_args.rdwr.expire, >))
			cp->fn_args.rdwr.expire	= now;

		timersub(&cp->fn_args.rdwr.expire,&now,&timeout);

		if (0 != event_add(&t->ev,&timeout)) {
			nio->sys_errno	= errno;
			nio->our_errno	= err
					= NIO_ESYSTEM;

			goto abort;
		}

		t->ev_mode	= m;
		t->ev_pending	= 1;

		break;
	case NIO_ABORT:
abort:
		if (t->ev_pending) {
			/* Cannot check error here because epoll(2) and
			 * others might return an error if the descriptor
			 * was closed and had already implicitly dequeued
			 * the event. Such is a benign circumstance, and
			 * very common in some situations.
			 */
			(void)event_del(&t->ev);

			t->ev_pending	= 0;
		}

		if (cp) {
			if (t == &sd.nio->rd) {
				assert(0 == cp->fn_args.read.cb(sd.nio, cp->fn_args.read.dstpos, n, err, cp->fn_args.read.cb_arg));
			} else {
				assert(0 == cp->fn_args.write.cb(sd.nio, cp->fn_args.write.srcpos, n, err, cp->fn_args.write.cb_arg));
			}
		}

		break;
	default:
		assert(!"Bad IO mode");
	}

done:
	if (sd.nio) {
		assert(&sd == SLIST_FIRST(&t->note_stack));

		SLIST_REMOVE_HEAD(&t->note_stack, sle);
	}

	return /* void */;
} /* nio_rdwr_try() */


enum nio_errno nio_read(struct nio *nio, void *dst, size_t dstlen, nio_read_cb cb, void *arg, struct timeval *timeout) {
	struct nio_stack_data sd	= { .nio = nio, .fn_name = &nio_read };
	enum nio_errno err		= NIO_ESUCCESS;
	struct timeval now;

	if (SLIST_EMPTY(&sd.nio->rd.call_stack))
		sd.nio->rd.num_calls	= 0;

	SLIST_INSERT_HEAD(&sd.nio->rd.call_stack, &sd, sle);

	if (sd.nio->rd.call.fn_name) {
		sd.nio->our_errno	= err
					= NIO_EPENDING;

		goto done;
	}

	if (0 != gettimeofday(&now,0)) {
		sd.nio->sys_errno	= errno;
		sd.nio->our_errno	= err
					= NIO_ESYSTEM;

		goto done;
	}

	assert(!sd.nio->rd.next);

	sd.nio->rd.call.fn_name			= sd.fn_name;
	sd.nio->rd.call.fn_args.read.dst	= dst;
	sd.nio->rd.call.fn_args.read.dstlen	= dstlen;
	sd.nio->rd.call.fn_args.read.dstpos	= dst;
	sd.nio->rd.call.fn_args.read.cb		= cb;
	sd.nio->rd.call.fn_args.read.cb_arg	= arg;

	timeradd(&now,timeout,&sd.nio->rd.call.fn_args.read.expire);

	sd.nio->rd.next	= &sd.nio->rd.call;

#if 1	/* Pass down the stack? */
	if (SLIST_NEXT(&sd, sle) && SLIST_NEXT(&sd, sle)->fn_name == sd.fn_name)
		goto done;
#endif

	while (sd.nio && sd.nio->rd.next && sd.nio->rd.next->fn_name == sd.fn_name) {
		sd.nio->rd.next	= 0;	/* Unset; Just pointer to .call. */

		nio_rdwr_try(sd.nio, &sd.nio->rd, (++sd.nio->rd.num_calls > sd.nio->opts.recurse_max));
	}

done:
	if (sd.nio) {
		assert(&sd == SLIST_FIRST(&sd.nio->rd.call_stack));

		SLIST_REMOVE_HEAD(&sd.nio->rd.call_stack, sle);
	}

	return err;
} /* nio_read() */


enum nio_errno nio_write(struct nio *nio, const void *src, size_t srclen, nio_write_cb cb, void *arg, struct timeval *timeout) {
	struct nio_stack_data sd	= { .nio = nio, .fn_name = &nio_write };
	enum nio_errno err		= NIO_ESUCCESS;
	struct timeval now;

	if (SLIST_EMPTY(&sd.nio->wr.call_stack))
		sd.nio->wr.num_calls	= 0;

	SLIST_INSERT_HEAD(&sd.nio->wr.call_stack, &sd, sle);

	if (sd.nio->wr.call.fn_name) {
		sd.nio->our_errno	= err
					= NIO_EPENDING;

		goto done;
	}

	if (0 != gettimeofday(&now,0)) {
		sd.nio->sys_errno	= errno;
		sd.nio->our_errno	= err
					= NIO_ESYSTEM;

		goto done;
	}

	assert(!sd.nio->wr.next);

	sd.nio->wr.call.fn_name			= sd.fn_name;
	sd.nio->wr.call.fn_args.write.src	= src;
	sd.nio->wr.call.fn_args.write.srclen	= srclen;
	sd.nio->wr.call.fn_args.write.srcpos	= src;
	sd.nio->wr.call.fn_args.write.cb	= cb;
	sd.nio->wr.call.fn_args.write.cb_arg	= arg;

	timeradd(&now,timeout,&sd.nio->wr.call.fn_args.read.expire);

	sd.nio->wr.next	= &sd.nio->wr.call;

#if 1	/* Pass down the stack? */
	if (SLIST_NEXT(&sd, sle) && SLIST_NEXT(&sd, sle)->fn_name == sd.fn_name)
		goto done;
#endif

	while (sd.nio && sd.nio->wr.next && sd.nio->wr.next->fn_name == sd.fn_name) {
		sd.nio->wr.next	= 0;	/* Unset; Just pointer to .call. */

		nio_rdwr_try(sd.nio, &sd.nio->wr, (++sd.nio->wr.num_calls > sd.nio->opts.recurse_max));
	}

done:
	if (sd.nio) {
		assert(&sd == SLIST_FIRST(&sd.nio->wr.call_stack));

		SLIST_REMOVE_HEAD(&sd.nio->wr.call_stack, sle);
	}

	return err;
} /* nio_write() */


static int nio_getln_tell(struct nio *nio, char *eol, enum nio_errno err) {
	struct nio_stack_data sd	= { .nio = nio, .fn_name = &nio_getln_tell };
	struct nio_call *cp	= &nio->gl.call;
	int again, ch;

	SLIST_INSERT_HEAD(&sd.nio->gl.note_stack, &sd, sle);

	if (cp->fn_args.getln.buf.pos > cp->fn_args.getln.buf.top) {
		assert(eol >= cp->fn_args.getln.buf.top && eol < cp->fn_args.getln.buf.end);

		ch	= *eol;
		*eol	= '\0';
	}

	again	= cp->fn_args.getln.cb(nio, cp->fn_args.getln.buf.top, eol - cp->fn_args.getln.buf.top, err, cp->fn_args.getln.cb_arg);

	if (sd.nio) {
		if (cp->fn_args.getln.buf.pos > cp->fn_args.getln.buf.top) {
			assert(eol >= cp->fn_args.getln.buf.top && eol < cp->fn_args.getln.buf.end);

			*eol	= ch;
		}

		(void)memmove(cp->fn_args.getln.buf.top, eol, cp->fn_args.getln.buf.pos - eol);

		cp->fn_args.getln.buf.pos	= cp->fn_args.getln.buf.top + (cp->fn_args.getln.buf.pos - eol);
	}

	if (sd.nio) {
		assert(&sd == SLIST_FIRST(&sd.nio->gl.note_stack));

		SLIST_REMOVE_HEAD(&sd.nio->gl.note_stack, sle);
	}

	return again;
} /* nio_getln_tell() */


static enum nio_action nio_getln_catch(struct nio *nio, void *bytes, size_t nbytes, enum nio_errno err, void *nul) {
	struct nio_call *cp	= &nio->gl.call;

	if (err != NIO_ESUCCESS) {
		cp->fn_name	= 0;

		if (err == NIO_ENOMORE && cp->fn_args.getln.buf.pos > cp->fn_args.getln.buf.top) {
			assert(0 == nio_getln_tell(nio, cp->fn_args.getln.buf.pos, err));
		} else {
			assert(0 == cp->fn_args.getln.cb(nio, 0, 0, err, cp->fn_args.getln.cb_arg));
		}

		return 0;
	}

	cp->fn_args.getln.buf.pos	+= nbytes;

	/*
	 * Call back into the public interface as-if starting from scratch
	 * so that we can trigger call coalescing.
	 */
	cp->fn_name	= 0;	/* Cheat. This is stupid, BTW. */
	nio_getln(nio, cp->fn_args.getln.cb, cp->fn_args.getln.cb_arg, &cp->fn_args.getln.timeout);

	return 0;
} /* nio_getln_catch() */


static void nio_getln_try(struct nio *nio) {
	struct nio_stack_data sd	= { .nio = nio, .fn_name = &nio_getln_try };
	struct nio_call *cp	= &sd.nio->gl.call;
	void *fn_name			= cp->fn_name;
	int again			= 1;
	char *nl;

	SLIST_INSERT_HEAD(&sd.nio->gl.note_stack, &sd, sle);

	while (again && (nl = memchr(cp->fn_args.getln.buf.top, '\n', cp->fn_args.getln.buf.pos - cp->fn_args.getln.buf.top))) {
		assert(sd.nio);

		cp->fn_name	= 0;
		again		= nio_getln_tell(nio, nl + 1, NIO_ESUCCESS);
	} /* while (again && nl) */

	if (again) {
		assert(sd.nio);

		cp->fn_name	= fn_name;

		if (0 == (cp->fn_args.getln.buf.end - cp->fn_args.getln.buf.pos)) {
			size_t want	= MAX(512, 2 * (cp->fn_args.getln.buf.end - cp->fn_args.getln.buf.top));
			char *p;

			if (!(p = (nio->ap->realloc)(nio->ap, cp->fn_args.getln.buf.top, want, 1)))
				goto sysfail;

			cp->fn_args.getln.buf.end	= p + want;
			cp->fn_args.getln.buf.pos	= p + (cp->fn_args.getln.buf.pos - cp->fn_args.getln.buf.top);
			cp->fn_args.getln.buf.top	= p;
		}

		if (0 != nio_read(nio, cp->fn_args.getln.buf.pos, (cp->fn_args.getln.buf.end - cp->fn_args.getln.buf.pos) - 1, &nio_getln_catch, 0, &cp->fn_args.getln.timeout))
				goto niofail;
	}

	goto done;
sysfail:
	nio->sys_errno	= errno;
	nio->our_errno	= NIO_ESYSTEM;

	/* FALL THROUGH */
niofail:
	cp->fn_name	= 0;

	assert(0 == cp->fn_args.getln.cb(nio, 0, 0, nio->our_errno, cp->fn_args.getln.cb_arg));

	/* FALL THROUGH */
done:
	if (sd.nio) {
		assert(&sd == SLIST_FIRST(&sd.nio->gl.note_stack));

		SLIST_REMOVE_HEAD(&sd.nio->gl.note_stack, sle);
	}

	return /* void */;
} /* nio_getln_try() */


enum nio_errno nio_getln(struct nio *nio, nio_getln_cb cb, void *arg, struct timeval *timeout) {
	struct nio_stack_data sd	= { .nio = nio, .fn_name = &nio_getln };
	enum nio_errno err		= NIO_ESUCCESS;
	struct timeval now;

	if (SLIST_EMPTY(&sd.nio->gl.call_stack))
		sd.nio->gl.num_calls	= 0;

	SLIST_INSERT_HEAD(&sd.nio->gl.call_stack, &sd, sle);

	if (sd.nio->gl.call.fn_name) {
		sd.nio->our_errno	= err
					= NIO_EPENDING;
assert(0);
		goto done;
	}

	if (0 != gettimeofday(&now,0)) {
		sd.nio->sys_errno	= errno;
		sd.nio->our_errno	= err
					= NIO_ESYSTEM;

		goto done;
	}

	assert(!sd.nio->gl.next);

	sd.nio->gl.call.fn_name			= sd.fn_name;
	sd.nio->gl.call.fn_args.getln.cb	= cb;
	sd.nio->gl.call.fn_args.getln.cb_arg	= arg;

	if (timeout) {
		sd.nio->gl.call.fn_args.getln.timeout	= *timeout;

		timeradd(&now,timeout,&sd.nio->gl.call.fn_args.getln.expire);
	}

	sd.nio->gl.next	= &sd.nio->gl.call;

	/* Pass down the stack? */
	if (SLIST_NEXT(&sd, sle) && SLIST_NEXT(&sd, sle)->fn_name == sd.fn_name)
		goto done;

	while (sd.nio && sd.nio->gl.next && sd.nio->gl.next->fn_name == sd.fn_name) {
		sd.nio->gl.next	= 0;	/* Unset; Just pointer to .call. */

		sd.nio->gl.num_calls++;

		nio_getln_try(sd.nio);
	}

done:
	if (sd.nio) {
		assert(&sd == SLIST_FIRST(&sd.nio->gl.call_stack));

		SLIST_REMOVE_HEAD(&sd.nio->gl.call_stack, sle);
	}

	return err;
} /* nio_getln() */


static enum nio_mode nio_sys_mode(struct nio *nio, int fd, enum nio_mode want, void *null) {
	return want;
} /* nio_sys_mode() */


static ssize_t nio_sys_read(struct nio *nio, int fd, void *p, size_t n, void *null) {
	return read(fd,p,n);
} /* nio_sys_read() */


static ssize_t nio_sys_write(struct nio *nio, int fd, const void *p, size_t n, void *null) {
	return write(fd,p,n);
} /* nio_sys_write() */


const struct nio_options nio_defaults = {
	.recurse_max	= -1,
	.sys_mode	= &nio_sys_mode,
	.sys_read	= &nio_sys_read,
	.sys_write	= &nio_sys_write,
};


struct nio *nio_open(int fd, const struct nio_options *opts, struct event_base *ev_base, const struct arena_prototype *ap) {
	struct nio *nio;

	if (!opts)
		opts	= &nio_defaults;

	if (!ap)
		ap	= ARENA_STDLIB;

	if (!(nio = ap->malloc(ap, sizeof *nio, 0)))
		return 0;

	*nio	= nio_initializer;

	nio->ap		= ap;
	nio->ev_base	= ev_base;
	nio->opts	= *opts;
	nio->fd		= fd;

	SLIST_INIT(&nio->rd.call_stack);
	SLIST_INIT(&nio->rd.note_stack);

	SLIST_INIT(&nio->wr.call_stack);
	SLIST_INIT(&nio->wr.note_stack);

	SLIST_INIT(&nio->gl.call_stack);
	SLIST_INIT(&nio->gl.note_stack);

	nio->gl.call.fn_args.getln.buf.top	= 0;
	nio->gl.call.fn_args.getln.buf.pos	= 0;
	nio->gl.call.fn_args.getln.buf.end	= 0;

	if (nio->opts.recurse_max == -1)
		nio->opts.recurse_max	= nio_recurse_max;

	return nio;
} /* nio_open() */


enum nio_errno nio_close(struct nio *nio) {
	struct nio_stack_data *sd;

	/* TODO: Allow early termination. */
	if (nio->wr.ev_pending || nio->rd.ev_pending)
		return NIO_EPENDING;

	/*
	 * Cleanup our call stack so we don't do stupid stuff (and if we do
	 * we [hopefully] immediately segfault).
	 */
	while ((sd = SLIST_FIRST(&nio->rd.call_stack))) {
		sd->nio	= 0;

		SLIST_REMOVE_HEAD(&nio->rd.call_stack, sle);
	}

	while ((sd = SLIST_FIRST(&nio->rd.note_stack))) {
		sd->nio	= 0;

		SLIST_REMOVE_HEAD(&nio->rd.note_stack, sle);
	}

	while ((sd = SLIST_FIRST(&nio->wr.call_stack))) {
		sd->nio	= 0;

		SLIST_REMOVE_HEAD(&nio->wr.call_stack, sle);
	}

	while ((sd = SLIST_FIRST(&nio->wr.note_stack))) {
		sd->nio	= 0;

		SLIST_REMOVE_HEAD(&nio->wr.note_stack, sle);
	}

	while ((sd = SLIST_FIRST(&nio->gl.call_stack))) {
		sd->nio	= 0;

		SLIST_REMOVE_HEAD(&nio->gl.call_stack, sle);
	}

	while ((sd = SLIST_FIRST(&nio->gl.note_stack))) {
		sd->nio	= 0;

		SLIST_REMOVE_HEAD(&nio->gl.note_stack, sle);
	}

	nio->ap->free(nio->ap,nio);

	return NIO_ESUCCESS;
} /* nio_close() */


int nio_sys_errno(struct nio *nio) {
	return nio->sys_errno;
} /* nio_sys_errno() */


enum nio_errno nio_errno(struct nio *nio) {
	return nio->our_errno;
} /* nio_errno() */


const char *nio_strerror(struct nio *nio) {
	if (nio->our_errno == NIO_ESYSTEM)
		return strerror(nio->sys_errno);
	else
		return nio_errlist[nio->our_errno];
} /* nio_strerror() */


