/* ==========================================================================
 * libevnet/src/socket.c - Network server library for libevent.
 * --------------------------------------------------------------------------
 * Copyright (c) 2006  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */
#include <stdio.h>	/* snprintf(3) */
#include <stdlib.h>	/* malloc(3) free(3) */
#include <stddef.h>	/* offsetof */

#include <errno.h>	/* EINVAL ESOCKNTNOSUPPORT EINPROGRESS EINTR
			 * EAGAIN ECONNABORTED */

#include <string.h>	/* strerror(3) memcpy(3) strncpy(3) */

#include <assert.h>	/* assert(3) */

#include <fcntl.h>	/* O_NONBLOCK F_SETFL F_GETFL fcntl(2) */

#include <sys/types.h>	/* socklen_t mode_t */
#include <sys/param.h>	/* MIN */
#include <sys/queue.h>	/* SLIST_HEAD SLIST_ENTRY SLIST_INIT SLIST_FIRST
			 * SLIST_NEXT LIST_HEAD LIST_ENTRY LIST_INIT
			 * LIST_INSERT LIST_REMOVE */
#include <sys/time.h>	/* struct timeval gettimeofday(2) timeradd(3)
			 * timercmp(3) timersub(3) */
#include <sys/stat.h>	/* umask(2) */
#include <sys/socket.h>	/* SOMAXCONN AF_UNIX AF_INET AF_INET6 SOCK_STREAM
			 * PF_UNIX PF_INET PF_INET6 SOL_SOCKET SO_ERROR
			 * struct sockaddr socket(2) connect(2) getsockopt(2) */

#include <netinet/in.h>	/* struct sockaddr_in struct sockaddr_in6 */
#include <sys/un.h>	/* struct sockaddr_un */

#include <unistd.h>	/* close(2) */

#include <netdb.h>	/* struct addrinfo struct servent getaddrinfo(3)
			 * getservbyname(3) */

#include <event.h>	/* EV_READ EV_WRITE EV_PERSIST EV_TIMEOUT
			 * struct event_base event_set(3) event_set_base(3)
			 * event_add(3) */

#include <arena/proto.h>

#include "socket.h"
#include "lookup.h"

#ifndef arena_alignof
#define arena_alignof(t)	offsetof(struct { char c; t x; }, x)
#endif


const char *socket_errlist[] = {
	[SOCKET_ESUCCESS]	= "Success",
	[SOCKET_ESYSTEM]	= "System error",
	[SOCKET_ETIMEDOUT]	= "Operation timed out",
	[SOCKET_ECANCELLED]	= "Operation cancelled",
	[SOCKET_ENOTFOUND]	= "Hostname not found",
}; /* socket_errlist[] */

const int socket_nerr	= sizeof socket_errlist / sizeof *socket_errlist;


const struct socket_options socket_defaults = {
	.cb_onclose		= 1,
	.sun_mask		= 0000,
	.sun_unlink		= 0,
	.sa_reuseaddr		= 0,
	.so_type		= SOCK_STREAM,
	.so_backlog		= SOMAXCONN,
	.so_nonblock		= 1,
}; /* socket_defaults */


struct stack_note {
	struct socket_service **sp;
	struct socket_accept **ap;
	struct socket_connect **cp;

	SLIST_ENTRY(stack_note) sle;
}; /* struct stack_note */

SLIST_HEAD(call_stack, stack_note);


static const struct socket_accept {
	int socket;

	struct call_stack stack;

	struct sockaddr_storage ss;
	socklen_t salen;

	struct event ev;
	int ev_pending;
	struct timeval ev_expires;

	socket_accept_cb cb;
	void *cb_arg;

	struct timeval timeout;
	int timeout_set;

	struct socket_service *service;

	LIST_ENTRY(socket_accept) le;
} socket_accept_initializer = {
	.socket	= -1,
};


static const struct socket_connect {
	int socket;

	struct call_stack stack;

	struct event ev;
	int ev_pending;

	struct lookup *lu;
	int lu_pending;

	socket_connect_cb cb;
	void *cb_arg;

	struct timeval expire;
	int expire_set;

	char qname[256];
	int qtype;
	unsigned short port;

	size_t num_ss;
	struct sockaddr_storage *ss_pos;
	struct sockaddr_storage *ss_buf;

	struct socket_service *service;

	LIST_ENTRY(socket_connect) le;
} socket_connect_initializer = {
	.socket	= -1, 
};


struct socket_service {
	struct socket_options opts;

	const struct arena_prototype *ap;
	struct event_base *ev_base;

	int sys_errno;
	int our_errno;

	struct call_stack stack;

	struct {
		LIST_HEAD(, socket_accept) listen;
	} accept;

	struct {
		LIST_HEAD(, socket_connect) pending;
	} connect;
} socket_service_initializer;


static int so_set_nonblock(int fd) {
	int flags;

	if (-1 == (flags = fcntl(fd, F_GETFL, 0)))
		return -1;

	if (0 != fcntl(fd, F_SETFL, flags | O_NONBLOCK))
		return -1;

	return 0;
} /* so_set_nonblock() */


static int so_open(struct sockaddr_storage *ss, int so_type) {
	int sd	= -1;

	switch (ss->ss_family) {
	case AF_INET:
		sd	= socket(PF_INET, so_type, 0);

		break;
	case AF_INET6:
		sd	= socket(PF_INET6, so_type, 0);

		break;
	case AF_UNIX:
		sd	= socket(PF_UNIX, so_type, 0);

		break;
	default:
		errno	= ESOCKTNOSUPPORT;
	} /* switch (ss_family) */

	return sd;
} /* so_open() */


struct socket_service *socket_open(const struct socket_options *opts, struct event_base *ev_base, const struct arena_prototype *ap) {
	struct socket_service *s	= 0;
	int sys_errno;

	if (!opts)
		opts	= &socket_defaults;

	if (!ap)
		ap	= ARENA_STDLIB;

	if (!(s = (ap->malloc)(ap, sizeof *s, arena_alignof(typeof(*s)))))
		goto sysfail;

	*s		= socket_service_initializer;
	s->opts		= *opts;
	s->ap		= ap;
	s->ev_base	= ev_base;

	SLIST_INIT(&s->stack);

	LIST_INIT(&s->accept.listen);
	LIST_INIT(&s->connect.pending);

	return s;
sysfail:
	sys_errno	= errno;

	(void)socket_close(s);

	errno		= sys_errno;

	return 0;
} /* socket_open() */


void socket_accept_close(struct socket_service *s, struct socket_accept *a) {
	struct stack_note *snp;

	assert(s != 0 && s == a->service);

	LIST_REMOVE(a, le);

	SLIST_FOREACH(snp, &a->stack, sle) {
		*snp->ap	= 0;
	}

	if (a->ev_pending)
		(void)event_del(&a->ev), a->ev_pending = 0;

	if (a->socket != -1)
		(void)close(a->socket), a->socket = -1;

	if (s->opts.cb_onclose && a->cb != 0) {
		socket_accept_cb cb	= a->cb;
		void *cb_arg		= a->cb_arg;

		memset(a, '\0', sizeof *a);

		(s->ap->free)(s->ap, a);

		cb(-1, 0, 0, SOCKET_ECANCELLED, cb_arg);
	} else {
		memset(a, '\0', sizeof *a);

		(s->ap->free)(s->ap, a);
	}

	return /* void */;
} /* socket_accept_close() */


void socket_connect_close(struct socket_service *s, struct socket_connect *c) {
	struct stack_note *snp;

	assert(s != 0 && s == c->service);

	LIST_REMOVE(c, le);

	SLIST_FOREACH(snp, &c->stack, sle) {
		*snp->cp	= 0;
	}

	if (c->ev_pending)
		(void)event_del(&c->ev), c->ev_pending = 0;

	if (c->socket != -1)
		(void)close(c->socket), c->socket = -1;

	/* TODO: Cancel the lookup when we support lookups. */
	assert(!c->lu_pending);

	if (s->opts.cb_onclose && c->cb != 0) {
		socket_connect_cb cb	= c->cb;
		void *cb_arg		= c->cb_arg;

		memset(c, '\0', sizeof *c);

		(s->ap->free)(s->ap, c);

		cb(-1, 0, 0, SOCKET_ECANCELLED, cb_arg);
	} else {
		memset(c, '\0', sizeof *c);

		(s->ap->free)(s->ap, c);
	}

	return /* void */;
} /* socket_connect_close() */


void socket_close(struct socket_service *s) {
	struct stack_note s_note	= { .sp	= &s, };
	struct stack_note *snp;
	struct socket_accept *a;
	struct socket_connect *c;

	if (!s)
		return /* void */;

	SLIST_FOREACH(snp, &s->stack, sle) {
		*snp->sp	= 0;
	}

	SLIST_INIT(&s->stack);

	SLIST_INSERT_HEAD(&s->stack, &s_note, sle);

	while (s != 0 && (a = LIST_FIRST(&s->accept.listen))) {
		socket_accept_close(s, a);
	} /* while (a = s->connect.accept) */

	while (s != 0 && (c = LIST_FIRST(&s->connect.pending))) {
		socket_connect_close(s, c);
	} /* while (c = s->connect.pending) */

	if (!s)
		return /* void */;

	assert(&s_note == SLIST_FIRST(&s->stack));

	SLIST_REMOVE_HEAD(&s->stack, sle);

	(s->ap->free)(s->ap, s);

	return /* void */;
} /* socket_close() */


static void socket_accept_catch(int, short, void *);

static int socket_accept_install(struct socket_accept *a, int restart, struct timeval *now) {
	struct timeval now_buf, timeout, expires;

	/* Short circuit if we're already pending without a timeout. */
	if (a->ev_pending && !a->timeout_set)
		return 0;

	if (!now) {
		if (0 != gettimeofday(&now_buf, 0))
			return -1;

		now	= &now_buf;
	}

	if (a->timeout_set) {
		if (restart || !a->ev_pending) {
			timeout	= a->timeout;	
		} else {
			if (timercmp(&a->ev_expires, now, <=)) {
				/* Oops. We passed it by! */
				timeout.tv_sec	= 0;
				timeout.tv_usec	= 0;
			} else {
				/* Take the difference. */
				timersub(&a->ev_expires, now, &timeout);
			}
		}

		timeradd(now, &timeout, &expires);

		/* Short circuit if the expiration is set properly. */
		if (a->ev_pending && timercmp(&expires, &a->ev_expires, ==))
			return 0;
	} /* if (a->timeout_set) */

	if (a->ev_pending)
		(void)event_del(&a->ev), a->ev_pending = 0;

	event_set(&a->ev, a->socket, EV_READ | EV_PERSIST, socket_accept_catch, a);

	if (a->service->ev_base)
		event_base_set(a->service->ev_base, &a->ev);

	if (0 != event_add(&a->ev, (a->timeout_set)? &timeout : 0))
		return -1;

	a->ev_pending	= 1;

	return 0;
} /* socket_accept_install() */


static void socket_accept_catch(int fd, short events, void *arg) {
	struct socket_accept *a		= arg;
	struct stack_note s_note	= { .ap = &a};
	unsigned int naccepted		= 0;
	struct timeval now;
	struct sockaddr_storage ss;
	socklen_t salen;

	/* After this, must always `goto cleanup' to return properly. */
	SLIST_INSERT_HEAD(&a->stack, &s_note, sle);

	/* Grab our entry time for stable time calculations. */
	if (0 != gettimeofday(&now, 0))
		goto sysfail;

	if (events & EV_TIMEOUT) {
		goto timeout;
	} else if (a->timeout_set){
		if (timercmp(&now, &a->ev_expires, >=))
			goto timeout;
	}

try_accept:
	salen	= sizeof ss;

	if (-1 == (fd = accept(a->socket, (struct sockaddr *)&ss, &salen))) {
		switch (errno) {
		case EINTR:
			/* FALL THROUGH */
		case EAGAIN:
			/* FALL THROUGH */
		case ECONNABORTED:
			/* FALL THROUGH */

			/* Recalculate any timeout on event entry. */
			if (0 != socket_accept_install(a, 0, &now))
				goto sysfail;

			/* NOTE: EV_PERSIST passed to every event_set(). */

			goto cleanup;
		default:
			goto sysfail;
		}
	}

	if (a->service->opts.so_nonblock && 0 != so_set_nonblock(fd))
		goto sysfail;

	/*
	 * Since EV_PERSIST is always set, adjust our time keep to
	 * approximate the next libevent expiration.
	 */
	if (a->timeout_set)
		timeradd(&now, &a->timeout, &a->ev_expires);

	/* Recalculate any timeout on event entry. */
	if (0 != socket_accept_install(a, 1, &now))
		goto sysfail;

	a->cb(fd, (struct sockaddr *)&ss, salen, SOCKET_ESUCCESS, a->cb_arg);

	/*
	 * Try to pull more than just one off the queue.
	 */
	if (a != 0 && naccepted++ < 3)
		goto try_accept;

	goto cleanup;
timeout:
	a->service->our_errno	= SOCKET_ETIMEDOUT;
	a->service->sys_errno	= ETIMEDOUT;

	/*
	 * Since EV_PERSIST is always set, adjust our time keep to
	 * approximate the next libevent expiration.
	 */
	if (a->timeout_set)
		timeradd(&now, &a->timeout, &a->ev_expires);

	/* Recalculate any timeout on event entry. */
	if (0 != socket_accept_install(a, 1, &now))
		goto sysfail;

	errno	= a->service->sys_errno;

	a->cb(-1, 0, 0, SOCKET_ETIMEDOUT, a->cb_arg);

	goto cleanup;
sysfail:
	a->service->our_errno	= SOCKET_ESYSTEM;
	a->service->sys_errno	= errno;

	if (a->ev_pending)
		(void)event_del(&a->ev), a->ev_pending = 0;

	errno	= a->service->sys_errno;

	a->cb(-1, 0, 0, SOCKET_ESYSTEM, a->cb_arg);

	goto cleanup;
cleanup:
	if (a) {
		assert(&s_note == SLIST_FIRST(&a->stack));

		SLIST_REMOVE_HEAD(&a->stack, sle);
	}

	return /* void */;
} /* socket_accept_catch() */


enum socket_errno socket_accept(struct socket_accept **ap, struct socket_service *s, struct socket_name *n, socket_accept_cb cb, void *cb_arg, struct timeval *timeout) {
	struct socket_accept *ap_buf, *a	= 0;
	mode_t omask;

	if (!ap)
		ap	= &ap_buf;

	*ap	= 0;

	if (!(a = (s->ap->malloc)(s->ap, sizeof *a, arena_alignof(typeof(*a)))))
		goto sysfail;

	*a		= socket_accept_initializer;
	a->service	= s;

	SLIST_INIT(&a->stack);

	/*
	 * Turn the struct socket_name into a struct sockaddr.
	 */
	if (n->type == &n->host) {
		errno	= EINVAL;

		goto sysfail;
	} else {
		(void)memset(&a->ss, '\0', sizeof a->ss);

		if (n->type == &n->local) {
			a->ss.ss_family	= AF_UNIX;

			assert(sizeof SOCKET_UN_P(&a->ss)->sun_path == sizeof n->local.path);

			(void)memcpy(SOCKET_UN_P(&a->ss)->sun_path, n->local.path, sizeof SOCKET_UN_P(&a->ss)->sun_path);
		} else if (n->type == &n->addr) {
			(void)memcpy(&a->ss, &n->addr.ss, sizeof a->ss);
		} else
			assert(0);

		a->salen	= SOCKET_SA_LENOF(&a->ss);
	}

	if (-1 == (a->socket = so_open(&a->ss, s->opts.so_type)))
		goto sysfail;

	if (s->opts.sa_reuseaddr
	&&  0 != setsockopt(a->socket, SOL_SOCKET, SO_REUSEADDR, (struct sockaddr *)&a->ss, a->salen))
		goto sysfail;

	if (0 != so_set_nonblock(a->socket))
		goto sysfail;

	if (a->ss.ss_family == AF_UNIX && s->opts.sun_unlink) {
		char path[sizeof (((struct sockaddr_un *)&a->ss)->sun_path) + 1];

		path[sizeof path - 1]	= '\0';

		/* Should we fail if errno != ENOENT? */
		(void)unlink(strncpy(path, ((struct sockaddr_un *)&a->ss)->sun_path, sizeof path - 1));
	}

	omask	= umask(s->opts.sun_mask);

	if (0 != bind(a->socket, (struct sockaddr *)&a->ss, a->salen)) {
		(void)umask(omask);

		goto sysfail;
	}

	(void)umask(omask);

	if (0 != listen(a->socket, s->opts.so_backlog))
		goto sysfail;

	if (timeout) {
		a->timeout	= *timeout;
		a->timeout_set	= 1;
	}
	
	a->cb		= cb;
	a->cb_arg	= cb_arg;

	if (0 != socket_accept_install(a, 1, 0))
		goto sysfail;

	LIST_INSERT_HEAD(&s->accept.listen, a, le);

	*ap	= a;

	return SOCKET_ESUCCESS;
sysfail:
	s->our_errno	= SOCKET_ESYSTEM;

	s->sys_errno	= errno;

	if (a != 0) {
		if (a->ev_pending)
			(void)event_del(&a->ev), a->ev_pending = 0;

		if (a->socket != -1)
			(void)close(a->socket), a->socket = -1;

		(s->ap->free)(s->ap, a);
	}

	errno		= s->sys_errno;

	return SOCKET_ESYSTEM;
} /* socket_accept() */


static void socket_connect_setup(struct socket_connect *);

static void socket_connect_catch(int fd, short event, void *arg) {
	struct socket_connect *c	= arg;
	struct socket_service *s	= c->service;
	enum socket_errno so_errno;
	struct timeval now, timeout;
	struct sockaddr_storage ss;
	struct sockaddr *sa		= 0;
	socklen_t salen			= 0;
	socket_connect_cb cb;
	void *cb_arg;
	int sd, sys_errno;

	if (event & EV_TIMEOUT)
		goto timeout;

	if (c->ev_pending) {
		int so_errno;
		socklen_t n	= sizeof so_errno;

		if (0 != getsockopt(c->socket, SOL_SOCKET, SO_ERROR, &so_errno, &n))
			goto sysfail;

		if (0 != so_errno) {
			errno	= so_errno;

			goto sysfail;
		}
	} else {
		if (0 != connect(c->socket, SOCKET_SA_P(c->ss_pos), SOCKET_SA_LENOF(c->ss_pos))) {
			if (errno == EINPROGRESS) {
				goto poll;
			} else {
				goto sysfail;
			}
		}
	}

	memset(&ss, '\0', sizeof ss);

	sa	= (struct sockaddr *)&ss;
	salen	= sizeof ss;

	if (0 != getpeername(c->socket, sa, &salen)) {
		sa	= 0;
		salen	= 0;

		goto sysfail;
	}

	so_errno	= SOCKET_ESUCCESS;

	goto done;
poll:
	if (0 != gettimeofday(&now, 0))
		goto sysfail;

	if (c->expire_set) {
		if (timercmp(&now, &c->expire, >=))
			goto timeout;

		timersub(&c->expire, &now, &timeout);
	}

	event_set(&c->ev, c->socket, EV_WRITE, socket_connect_catch, c);

	if (0 != event_add(&c->ev, (c->expire_set)? &timeout : 0))
		goto sysfail;

	c->ev_pending	= 1;

	return /* void */;
timeout:
	if (c->socket != -1)
		(void)close(c->socket), c->socket = -1;

	so_errno	= SOCKET_ETIMEDOUT;

	goto done;
sysfail:
	so_errno	= SOCKET_ESYSTEM;

	sys_errno	= errno;

	if (c->socket != -1)
		(void)close(c->socket), c->socket = -1;

	errno		= sys_errno;

	if (c->ss_pos < &c->ss_buf[c->num_ss - 1]) {
		socket_connect_setup(c);

		return;
	}

	/* FALL THROUGH */
done:
	sd	= c->socket;
	cb	= c->cb;
	cb_arg	= c->cb_arg;

	LIST_REMOVE(c, le);

	c->service	= 0;

	sys_errno	= errno;

	(s->ap->free)(s->ap, c->ss_buf);
	(s->ap->free)(s->ap, c);

	errno		= sys_errno;

	cb(sd, sa, salen, so_errno, cb_arg);

	return /* void */;
} /* socket_connect_catch() */


static void socket_connect_setup(struct socket_connect *c) {
	struct socket_service *s	= c->service;
	int sys_errno;
	socket_connect_cb cb;
	void *cb_arg;

next:
	assert(c->ss_pos < &c->ss_buf[c->num_ss]);

	while (c->ss_pos < &c->ss_buf[c->num_ss]) {
		if (-1 != (c->socket = so_open(c->ss_pos, s->opts.so_type)))
			break;
		else if (c->ss_pos >= &c->ss_buf[c->num_ss - 1])
			break;
		else
			c->ss_pos++;
	}

	if (c->socket == -1)
		goto sysfail;

	if (0 != so_set_nonblock(c->socket))
		goto sysfail;

	socket_connect_catch(c->socket, EV_WRITE, c);

	return /* void */;
sysfail:
	sys_errno	= errno;

	if (c->socket != -1)
		(void)close(c->socket), c->socket = -1;

	if (c->ss_pos < &c->ss_buf[c->num_ss - 1])
		goto next;

	cb	= c->cb;
	cb_arg	= c->cb_arg;

	LIST_REMOVE(c, le);

	c->service	= 0;

	(s->ap->free)(s->ap, c->ss_buf);
	(s->ap->free)(s->ap, c);

	errno		= sys_errno;

	cb(-1, 0, 0, SOCKET_ESYSTEM, cb_arg);

	return /* void */;
} /* socket_connect_setup() */


void socket_connect_catch_lookup(int nans, struct lookup_rr *ans, int nadd, struct lookup_rr *add, enum lookup_errno errnum, void *arg) {
	struct socket_connect *c	= arg;
	socket_connect_cb cb		= c->cb;
	void *cb_arg			= c->cb_arg;
	int sys_errno			= errno;
	struct lookup_rr *r;

	if (LOOKUP_FAILURE(errno)) {
		goto dnsfail;
	} else if (!(ans > 0)) {
		sys_errno	= EFAULT;

		goto sysfail;
	} else if (!(c->ss_buf = c->service->ap->malloc(c->service->ap, nans * sizeof *c->ss_buf, 0))) {
		goto sysfail;
	}

	for (r = ans, c->ss_pos = c->ss_buf; r != 0; r = r->next) {
		switch (r->type) {
		case LOOKUP_IN_A:
			r->rr.ip.sa.sin.sin_port	= c->port;

			break;
		case LOOKUP_IN_AAAA:
			r->rr.ip.sa.sin6.sin6_port	= c->port;

			break;
		default:
			continue;
		}

		(void)memcpy(c->ss_pos, &r->rr.ip.sa, r->rr.ip.salen);
#if 0
		c->ss_len	= r->rr.ip.salen;
#endif

		c->ss_pos++;
	}

	c->num_ss	= c->ss_pos - c->ss_buf;
	c->ss_pos	= c->ss_buf;

	if (c->num_ss == 0) {
		sys_errno	= EFAULT;

		goto sysfail;
	}

	socket_connect_setup(c);

	return /* void */;
sysfail:
	errnum	= LOOKUP_ESYSTEM;

	/* FALL THROUGH */
dnsfail:
	(c->service->ap->free)(c->service->ap, c->ss_buf);
	(c->service->ap->free)(c->service->ap, c);

	errno		= sys_errno;

	switch (errnum) {
	case LOOKUP_ESYSTEM:
		cb(-1, 0, 0, SOCKET_ESYSTEM, cb_arg);
		break;
	case LOOKUP_ETIMEDOUT:
		cb(-1, 0, 0, SOCKET_ETIMEDOUT, cb_arg);
		break;
	default:
		cb(-1, 0, 0, SOCKET_ENOTFOUND, cb_arg);
	}

	return /* void */;
} /* socket_connect_catch_lookup() */


void socket_connect(struct socket_connect **cp, struct socket_service *s, struct socket_name *n, socket_connect_cb cb, void *cb_arg, struct timeval *timeout) {
	struct socket_connect *cp_buf, *c	= 0;
	struct timeval now;
	int sys_errno;

	if (!cp)
		cp	= &cp_buf;

	*cp	= 0;

	if (!(c = (s->ap->malloc)(s->ap, sizeof *c, arena_alignof(typeof(*c)))))
		goto sysfail;

	*c	= socket_connect_initializer;

	*cp	= c;

	/*
	 * Turn the struct socket_name into a struct sockaddr.
	 */
	if (n->type == &n->host) {
		struct addrinfo hints, *res, *r;

		(void)memset(&hints, 0, sizeof hints);

		hints.ai_flags		= AI_NUMERICHOST;
		hints.ai_family		= AF_UNSPEC;
		hints.ai_socktype	= s->opts.so_type;

		if (0 == getaddrinfo(n->host.name, n->host.port, &hints, &res)) {
			for (c->num_ss = 0, r = res; r != 0; c->num_ss++, r = r->ai_next)
				;;

			assert(c->num_ss > 0);

			if (!(c->ss_buf = (s->ap->malloc)(s->ap, sizeof *c->ss_buf * c->num_ss, arena_alignof(typeof(*c->ss_buf))))) {
				freeaddrinfo(res);

				goto sysfail;
			}

			for (c->ss_pos = c->ss_buf, r = res; r != 0; c->ss_pos++, r = r->ai_next)
				(void)memcpy(c->ss_pos, r->ai_addr, MIN(sizeof *c->ss_pos, r->ai_addrlen));

			c->ss_pos	= c->ss_buf;

			freeaddrinfo(res);
		} else {
			struct servent *ent;
			long port;
			char *end;

			port	= strtol(n->host.port, &end, 10);

			if (port < 0 || port > 65535 || *end != '\0') {
				if (!(ent = getservbyname(n->host.port, "tcp")))
					goto sysfail;

				c->port	= ent->s_port;
			} else
				c->port	= port;

			strncpy(c->qname, n->host.name, sizeof c->qname);
			c->qtype	= LOOKUP_IN_A | LOOKUP_IN_AAAA;
		}
	} else {
		if (!(c->ss_buf = (s->ap->malloc)(s->ap, sizeof *c->ss_buf, arena_alignof(typeof(*c->ss_buf)))))
			goto sysfail;

		c->num_ss	= 1;
		c->ss_pos	= c->ss_buf;

		if (n->type == &n->local) {
			(void)memset(c->ss_buf, '\0', sizeof *c->ss_buf);

			c->ss_buf->ss_family	= AF_UNIX;

			assert(sizeof SOCKET_UN_P(c->ss_buf)->sun_path == sizeof n->local.path);

			(void)memcpy(SOCKET_UN_P(c->ss_buf)->sun_path, n->local.path, sizeof SOCKET_UN_P(c->ss_buf)->sun_path);
		} else if (n->type == &n->addr) {
			(void)memcpy(c->ss_buf, &n->addr.ss, sizeof *c->ss_buf);
		} else
			assert(0);
	}

	c->cb		= cb;
	c->cb_arg	= cb_arg;

	if (timeout) {
		if (0 != gettimeofday(&now, 0))
			goto sysfail;

		timeradd(&now, timeout, &c->expire);

		c->expire_set = 1;
	}

	c->service	= s;

	LIST_INSERT_HEAD(&s->connect.pending, c, le);

	if (n->type == &n->host) {
		if (c->num_ss > 0)
			socket_connect_setup(c);
		else
			lookup(c->qname, strlen(c->qname), c->qtype, 0, &socket_connect_catch_lookup, c, timeout);
	} else {
		socket_connect_setup(c);
	}

	return /* void */;
sysfail:
	/*
	 * If we got here the structure wasn't inserted into the pending
	 * queue of the socket service context.
	 */
	sys_errno	= errno;

	if (c) {
		(s->ap->free)(s->ap, c->ss_buf);
		(s->ap->free)(s->ap, c);
	}

	errno		= sys_errno;

	*cp		= 0;

	cb(-1, 0, 0, SOCKET_ESYSTEM, cb_arg);

	return /* void */;
} /* socket_connect() */


struct socket_name *soname_set_addr(struct socket_name *dst, const struct sockaddr *sa, socklen_t salen) {
	dst->type	= &dst->addr;

	return memcpy(&dst->addr.sa, sa, salen);
} /* soname_set_addr */


struct socket_name *soname_set_local(struct socket_name *dst, const char *path) {
	dst->type	= &dst->local;

	(void)snprintf(dst->local.path, sizeof dst->local.path, "%s", path);

	return dst;
} /* soname_set_addr */


struct socket_name *soname_set_host(struct socket_name *dst, const char *name, const char *port) {
	dst->type	= &dst->host;

	(void)snprintf(dst->host.name, sizeof dst->host.name, "%s", name);
	(void)snprintf(dst->host.port, sizeof dst->host.port, "%s", port);

	return dst;
} /* soname_set_host */


enum socket_errno socket_errno(struct socket_service *s) {
	return s->our_errno;
} /* socket_errno() */


const char *socket_strerror(struct socket_service *s) {
	return (s->our_errno == SOCKET_ESYSTEM)?
		strerror(s->sys_errno) : socket_errlist[s->our_errno];
} /* socket_strerror() */

